#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include "gpio_dev.h"
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#include "gpiotimer.h"

void print_usage()
{
    printf("gpiotimer gpio\n");
    exit(0);
}

struct GpiotimerData data = { 0, 0.0, 0 };

pthread_mutex_t mutex;

void leave(int sig);
int fd_gpio = 0;
int thread_rc = 0;
int sock = 0;

void *SocketThread(void *_data)
{
    printf("Thread\n");

    struct sockaddr_in name;
    sock = socket(PF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
        perror("socket");
        exit(EXIT_FAILURE);
    }
    int val = 1;
    setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof( val ) );
    
    name.sin_family = AF_INET;
    name.sin_port = htons(3000);
    name.sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(sock, (struct sockaddr *) &name, sizeof(name)) < 0) {
        perror("bind");
        exit(EXIT_FAILURE);
    }

    if (listen(sock, 1) < 0) {
        perror("listen");
        exit(EXIT_FAILURE);
    }

    while (1) {
        int clientsock = accept(sock, NULL, NULL);
        if (clientsock < 0) {
            //Error
        } else {
            pthread_mutex_lock(&mutex);
            if (send(clientsock, &data, sizeof(struct GpiotimerData), 0) < 0) {
                //Error
                perror("send");
            }
            pthread_mutex_unlock(&mutex);

            close(clientsock);
        }
    }

}

int calcPin(const char *name)
{
    // Normal Number
    if(name[0]>='0' && name[0]<='9')
        return atoi(name);
    
    // Name
    if(name[0]=='P'){
        int pin=0;
        if(name[1]=='A')
            pin=32;
        else if(name[1]=='B')
            pin=64;
        else if(name[1]=='C')
            pin=96;
        else{
            fprintf(stderr,"Unknown Pin %s!\n",name);
            exit(EXIT_FAILURE);
        }
        
        pin += atoi(name+2);
        return pin;
    }
    fprintf(stderr,"Unknown Pin %s!\n",name);
    exit(EXIT_FAILURE);
}


int main(int argc, char **argv)
{
    int gpio_pin;

    int counter = 0;
    time_t dtime_all = 0;

    int counter2 = 0;
    time_t dtime_2 = 0;
    float last_W = 0.0;

    if (argc != 2) {
        print_usage();
    }

    if ((fd_gpio = open("/dev/gpio", O_RDWR)) < 0) {
        fprintf(stderr, "Error whilst opening /dev/gpio\n");
        return -1;
    }

    gpio_pin = calcPin(argv[1]);

    // Create Mutex
    pthread_mutex_init(&mutex, NULL);

    // Create Thread for socket
    pthread_t thread;
    thread_rc = pthread_create(&thread, NULL, SocketThread, NULL);
    if (thread_rc) {
        fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", thread_rc);
        close(fd_gpio);
        exit(-1);
    }
    // file and socket should be closed
    (void) signal(SIGINT, leave);

    printf("using gpio pin %d\n", gpio_pin);

    time_t begintime = time(NULL);
    time_t begintime2 = time(NULL);
    ioctl(fd_gpio, GPIO_REQUEST, gpio_pin);


    int oldval = ioctl(fd_gpio, GPIO_GET, gpio_pin);

    time_t oldtime = time(NULL);

    while (1) {
        int val = ioctl(fd_gpio, GPIO_GET, gpio_pin);
        int changed = val != oldval;
        oldval = val;
        if (changed) {
            //printf("pin %d changed to %s\n", gpio_pin, (val ? "HIGH" : "LOW"));
            if (!val)           // LOW
            {
                counter++;
                counter2++;
            }
        }

        time_t newtime = time(NULL);

        // Update data
        if (changed || newtime > oldtime) {
            oldtime = newtime;
            // Data for the last 10 seconds
            dtime_2 = newtime - begintime2;
            if (dtime_2 >= 10) {
                last_W = (float) (counter2 * 3600) / dtime_2;
                printf("%f W\n", last_W);
                begintime2 = newtime;
                counter2 = 0;
            }

            dtime_all = newtime - begintime;

            // Copy data for other thread
            pthread_mutex_lock(&mutex);
            data.counter = counter;
            data.last_W = last_W;
            data.dtime_all = dtime_all;
            pthread_mutex_unlock(&mutex);
        }

        usleep(5);
    }

    return 0;
}

void leave(int sig)
{
    printf("\nInterrupted..\n");
    close(fd_gpio);
    close(sock);
    exit(sig);
}
